#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <linux/fb.h>
#include <signal.h>
#include <time.h>

#include <rfb/rfb.h>

#define WINDOW_WIDTH	320
#define WINDOW_HEIGHT	240
#define WINDOW_DEPTH	2

#define FPS_COUNTER 1

/* from tslib.h */
struct TS_EVENT {
	unsigned short	pressure;
	unsigned short	x;
	unsigned short	y;
	unsigned short	pad;
} tsemu_sample;

/* device for emulation of touchscreen */
int tsemu_dev;

/* SIGPIPE signal handler : we catch SIGPIPE and ignore it so the library's
   write calls receive -EPIPE but the server keeps on running */
void sighandler(int signum)
{
	return;
}

void doptr(int buttonMask,int x,int y,rfbClientPtr cl)
{
	tsemu_sample.x=x;
	tsemu_sample.y=y;
	tsemu_sample.pressure=buttonMask;

	write(tsemu_dev, &tsemu_sample, sizeof(struct TS_EVENT));

	//fprintf(stderr,"Writing to /dev/touchscreen/wm97xx @%d,%d pressure %d<n",x,y,buttonMask);

	//rfbDefaultPtrAddEvent(buttonMask,x,y,cl);
}

int main(int argc,char** argv)
{
	int fbdev;
	unsigned short int *fb,*cmpb, *rfb;
	rfbScreenInfoPtr rfbScreen;

	int argi = 1,mode=1,usleepmicro=0;
	#ifdef FPS_COUNTER
	int showfps=0;
	#endif
	while(argi < argc){
		if(strcmp(argv[argi],"--full") == 0){
			mode = 1;
			fprintf(stderr, "DEBUG: Full mode\n");
		}else if(strcmp(argv[argi],"--none") == 0){
			mode = 0;
		#ifdef FPS_COUNTER
		}else if(strcmp(argv[argi],"--showfps") == 0){
			showfps = 1;
			fprintf(stderr, "DEBUG: Show FPS\n");
		#endif
		}else if(strcmp(argv[argi],"--usleep") == 0 && argi+1 < argc){
			usleepmicro = atoi(argv[argi+1]);
			argi++;
			fprintf(stderr, "DEBUG: Usleep %d\n",usleepmicro);
		}
		argi++;
	}

	fprintf(stderr,"GP2X VNC SERVER v0.9 by Lithosphere\n");
	fprintf(stderr," using LibVNCServer 0.7.1\n\n");

	/* Open /dev/fb0 */
	fbdev=open("/dev/fb0", O_RDONLY);
	if(fbdev < 0){
		fprintf(stderr, "DEBUG: Error opening /dev/fb0\n");
		return(-1);
	}

	/* Malloc compare buffer */
	cmpb = (unsigned short int*)malloc(WINDOW_WIDTH * WINDOW_HEIGHT * WINDOW_DEPTH);
	if(cmpb == NULL){
		close(fbdev);
		fprintf(stderr, "DEBUG: Error malloc compare buffer\n");
		return(-1);
	}

	/* Malloc remote buffer */
	rfb = (unsigned short int*)malloc(WINDOW_WIDTH * WINDOW_HEIGHT * WINDOW_DEPTH);
	if(rfb == NULL){
		close(fbdev);
		free(cmpb);
		fprintf(stderr, "DEBUG: Error malloc remote frame buffer\n");
		return(-1);
	}

	/* Mmap /dev/fb */
	fb=(unsigned short int *)mmap(0, WINDOW_WIDTH * WINDOW_HEIGHT * WINDOW_DEPTH, PROT_READ, MAP_SHARED, fbdev, 0);
	if(fb < 0){
		close(fbdev);
		free(cmpb);
		free(rfb);
		fprintf(stderr, "DEBUG: Error mmapping /dev/fb0\n");
		return(-1);
	}
	fprintf(stderr, "DEBUG: fb opened at %x, fd %d\n", fb, fbdev);

	/* Open tsemu /dev/tsemu */
	tsemu_dev=open("/dev/touchscreen/wm97xx", O_WRONLY);
	if(tsemu_dev < 0){
		fprintf(stderr, "DEBUG: Error opening /dev/touchscreen/wm97xx - Disabling touchscreen support\n");
		//return(-1);
	}else{
		fprintf(stderr, "DEBUG: tsemudev opened fd %d\n", tsemu_dev);
	}

	rfbScreen = rfbGetScreen(&argc, argv, WINDOW_WIDTH, WINDOW_HEIGHT, 5, 3,WINDOW_DEPTH);
	rfbScreen->desktopName = "GP2X VNC Server";
	rfbScreen->frameBuffer=(char*)rfb;
	if(tsemu_dev > 0)
		rfbScreen->ptrAddEvent = doptr;
	rfbInitServer(rfbScreen);

	/* Make sure we catch EPIPE so we don't get killed by closed sockets when
		the library tries to write to them */
	signal(SIGPIPE,sighandler);

	#ifdef FPS_COUNTER
	int numframes=0,startclock;
	#endif

	int i=0;

	while(1){
		while(rfbScreen->clientHead==NULL) {
			rfbProcessEvents(rfbScreen,1000);
		}
		/* Process events */
		for(i=0;i<20;i++)
			rfbProcessEvents(rfbScreen,1000);

		if(mode == 0){

		}else if(mode == 1){

		/* Full Copy mode, copy the whole screen */
			int pix;
			unsigned short int convert,convert2,convert3;
			for(pix=0;pix<WINDOW_WIDTH * WINDOW_HEIGHT;pix++){
				convert = (fb[pix] & (16+8+4+2+1)) << 10;
				convert2 = (fb[pix] & (1024+512+256+128+64)) >> 1;
				convert3 = (fb[pix] & (32768+16384+8192+4096+2048)) >> 11;

				rfb[pix] = convert | convert2 | convert3;
				//rfb[pix] = fb[pix];
			}
			rfbMarkRectAsModified(rfbScreen,0,0,WINDOW_WIDTH,WINDOW_HEIGHT);

		}else{

/* Here we check 2 pixels at once, and even if only one pixel changes we copy them both,
this speeds up checking for changed pixels. */

#define LOW_B ((fb2[pos] & (16+8+4+2+1)) << 10)
#define LOW_G ((fb2[pos] & (1024+512+256+128+64)) >> 1)
#define LOW_R ((fb2[pos] & (32768+16384+8192+4096+2048)) >> 11)
#define HIGH_B ((fb2[pos] & ((16+8+4+2+1)<<16)) << 10)
#define HIGH_G ((fb2[pos] & ((1024+512+256+128+64)<<16)) >> 1)
#define HIGH_R ((fb2[pos] & ((32768+16384+8192+4096+2048)<<16)) >> 11)

			/* Copy only the changed pixels */
			int pos,x1=WINDOW_WIDTH,y1=WINDOW_HEIGHT,x2=-1,y2=-1;
			unsigned int *fb2 = (unsigned int *)fb,*cmpb2 = (unsigned int *)cmpb,*rfb2 = (unsigned int *)rfb;
			for(pos=0;pos<WINDOW_WIDTH * WINDOW_HEIGHT/2;pos++){
				if(fb2[pos] != cmpb2[pos]){
					int tmp = pos%(WINDOW_WIDTH/2);

					if(x1 > tmp)
						x1 = tmp;

					if(x2 < tmp)
						x2 = tmp;

					tmp = pos/(WINDOW_WIDTH/2);
					if(y1 > tmp)
						y1 = tmp;

					if(y2 < tmp)
						y2 = tmp;

					cmpb2[pos] = fb2[pos];

					rfb2[pos] = HIGH_B | HIGH_G | HIGH_R | LOW_B | LOW_G | LOW_R;
				}
			}

			if(x2 != -1){
				rfbMarkRectAsModified(rfbScreen,x1<<1,y1,(x2<<1)+1,y2+1);
				//fprintf(stderr,"%d: RECT [%d, %d] x [%d, %d]\n",clock()/CLOCKS_PER_SEC,x1<<1,y1,(x2<<1)+1,y2);
			}//else
				//fprintf(stderr,"%d: Nothing changed\n",clock()/CLOCKS_PER_SEC);
		}


		// Sleep
		if(usleepmicro)
			usleep(usleepmicro);

#ifdef FPS_COUNTER
		if(numframes%10 == 0 && showfps){
			if(numframes == 10)
				fprintf(stderr,"FPS: %d\n",CLOCKS_PER_SEC/(clock()-startclock+1));	// fixed division by zero
			startclock=clock();
			numframes=0;
		}
		numframes++;
#endif

	}

	return(0);
}
